%==========================================================================
%
%   ALBATROS.m
%
%   Authors: SUMO Lab Team
%   Version: 1.0
%   Copyright: 2017
%   
%   When referring to ALBATROS please make reference to the corresponding publication:
%   
%   Van Steenkiste T., J. van der Herten, D. Deschrijver and T. Dhaene.
%   "ABLATROS: Adaptive Line-Based Sampling Trajectories for Sequential
%   Measurements". Engineering with Computers.
%
%==========================================================================
%
%   SIGNATURE:
%   [ wayPointsToTravel ] = ALBATROS( samples, boundingBox, stepSize, distanceFunction )
%
%==========================================================================
%
%   PURPOSE:
%   Determine new waypoints to travel on using the ALBATROS sequential
%   line-based sampling algorithm.
%
%   INPUT:
%   samples: previously collected samples
%   boundingBox: bounds of the space in which the algorithm should run
%   stepSize: suggested stepsize for the sampling
%   distanceFunction: function which computes a distance between two points
%
%   OUTPUT:
%   wayPointsToTravel: new points to travel on and gather samples
%==========================================================================

function [ wayPointsToTravel ] = ALBATROS( samples, boundingBox, stepSize, distanceFunction )
    %% store some initial values
    lastWayPoint = samples(end,:);
    numericalPrecision = 6;
    
    %% construct the Voronoi diagram
    voronoiDiagram = mpt_voronoi(samples','bound',Polyhedron(boundingBox));
    
    %% extract edges and vertices from Voronoi diagram
    % initialise values
    voronoiVertices = [];
    voronoiEdges = [];
    % loop over all Voronoi cells
    for cellIndex = 1:voronoiDiagram.Num
        voronoiCell = voronoiDiagram.Set(cellIndex).minHRep();
        % loop over all facets of Voronoi cell
        for facetIndex = 1:size(voronoiCell.H,1)
           facet = voronoiCell.getFacet(facetIndex);
           facetVertices = sortPointsClockwise(facet.V);
           voronoiVertices = [voronoiVertices; facetVertices];
           facetVertices = [facetVertices; facetVertices(1,:)];
           for vertexIndex = 1:(size(facetVertices,1)-1)
               voronoiEdges = [voronoiEdges;[facetVertices(vertexIndex,:) facetVertices(vertexIndex+1,:)]];
           end
        end
    end
    % only keep unique values
    voronoiVertices = unique(round(voronoiVertices,numericalPrecision),'rows');
    voronoiEdges = unique(round(voronoiEdges,numericalPrecision),'rows');
    voronoiEdgesTmp = cell(size(voronoiEdges,1),1);
    for edgeIndex = 1:length(voronoiEdgesTmp)
       if size(voronoiEdges,2) == 4
           % 2D edges
           voronoiEdgesTmp{edgeIndex} = struct('A',voronoiEdges(edgeIndex,1:2),'B',voronoiEdges(edgeIndex,3:4));
       elseif size(voronoiEdges,2) == 6
           % 3D edges
           voronoiEdgesTmp{edgeIndex} = struct('A',voronoiEdges(edgeIndex,1:3),'B',voronoiEdges(edgeIndex,4:6));
       end
    end
    voronoiEdges = voronoiEdgesTmp;
    
    %% determine next waypoint using maximin algorithm
    maximinDistances = ones(size(voronoiVertices,1),1) * Inf;
    for i = 1:size(voronoiVertices,1)
        for j = 1:size(samples,1)
            dist = distanceFunction(samples(j,:),voronoiVertices(i,:));
            if dist < maximinDistances(i)
                maximinDistances(i) = dist;
            end
        end
    end
    [~,idx] = max(maximinDistances);
    newWayPoint = voronoiVertices(idx,:);
    
    %% project the last waypoint to the closest voronoiEdge
    closestDistance = inf;
    for edgeIndex = 1:length(voronoiEdges)
        projection = projectPointToLine(voronoiEdges{edgeIndex}.A,voronoiEdges{edgeIndex}.B,lastWayPoint);
        distance = distanceFunction(lastWayPoint,projection);
        if distance < closestDistance
           closestDistance = distance;
           projectionLastWayPoint = projection;
           projectionLastWayPointA = voronoiEdges{edgeIndex}.A;
           projectionLastWayPointB = voronoiEdges{edgeIndex}.B;
           distanceProjectionLastWayPointToA = distanceFunction(projection,voronoiEdges{edgeIndex}.A);
           distanceProjectionLastWayPointToB = distanceFunction(projection,voronoiEdges{edgeIndex}.B);
        end
    end
    
    %% determine which of the Voronoi vertices is closest to the new waypoint
    closestDistance = inf;
    for vertexIndex = 1:size(voronoiVertices,1)
        distance = distanceFunction(newWayPoint,voronoiVertices(vertexIndex,:));
        if distance < closestDistance
            newWayPointClosestVertexDistance = distance;
            newWayPointClosestVertex = voronoiVertices(vertexIndex,:);
        end
    end
    
    %% add these new points as potential waypoints to the Voronoi vertices
    wayPoints = [voronoiVertices;projectionLastWayPoint;lastWayPoint;newWayPoint];
    
    %% compute the weights of the Voronoi edges
    edgeWeights = zeros(size(wayPoints,1));
    % loop over all cells in the Voronoi diagram
    for cellIndex = 1:voronoiDiagram.Num
       cellSamples = samples(voronoiDiagram.Set(cellIndex).contains(samples'),:);
       % loop over all facets of the cell
       for facetIndex = 1:size(voronoiDiagram.Set(cellIndex).H,1)
           facetVertices = sortPointsClockwise(voronoiDiagram.Set(cellIndex).getFacet(facetIndex).V);
           facetVertices = [facetVertices;facetVertices(1,:)];
           for vertexIndex = 1:(size(facetVertices,1)-1)
              A = facetVertices(vertexIndex,:);
              B = facetVertices(vertexIndex+1,:);
              [~,Aidx] = min(sum(abs(wayPoints - repmat(A,size(wayPoints,1),1)),2));
              [~,Bidx] = min(sum(abs(wayPoints - repmat(B,size(wayPoints,1),1)),2));
              weight = 0;
              for cellSampleIndex = 1:size(cellSamples,1)
                 cellSample = cellSamples(cellSampleIndex,:);
                 cellSampleProjection = projectPointToLine(A,B,cellSample);
                 weight = weight + (distanceFunction(A,B)*distanceFunction(cellSample,cellSampleProjection)) / 2;
              end
              edgeWeights(Aidx,Bidx) = edgeWeights(Aidx,Bidx) + weight;
              edgeWeights(Bidx,Aidx) = edgeWeights(Bidx,Aidx) + weight;
           end
       end
    end
    for edgeIndex = 1:length(voronoiEdges)
       [~,Aidx] =  min(sum(abs(wayPoints - repmat(voronoiEdges{edgeIndex}.A,size(wayPoints,1),1)),2));
       [~,Bidx] =  min(sum(abs(wayPoints - repmat(voronoiEdges{edgeIndex}.B,size(wayPoints,1),1)),2));
       edgeWeights(Aidx,Bidx) = edgeWeights(Aidx,Bidx) + 0.1;
       edgeWeights(Bidx,Aidx) = edgeWeights(Bidx,Aidx) + 0.1;
    end
    maxWeight = max(max(edgeWeights));
    zeroWeights = (edgeWeights == 0);
    edgeWeights = maxWeight + 1 - edgeWeights;
    edgeWeights(zeroWeights) = 0;
    
    %% add additional weights to matrix
    [~,lastWayPointIndex] = min(sum(abs(wayPoints - repmat(lastWayPoint,size(wayPoints,1),1)),2));
    [~,projectionLastWayPointIndex] = min(sum(abs(wayPoints - repmat(projectionLastWayPoint,size(wayPoints,1),1)),2));
    [~,projectionLastWayPointAIndex] = min(sum(abs(wayPoints - repmat(projectionLastWayPointA,size(wayPoints,1),1)),2));
    [~,projectionLastWayPointBIndex] = min(sum(abs(wayPoints - repmat(projectionLastWayPointB,size(wayPoints,1),1)),2));
    [~,newWayPointIndex] = min(sum(abs(wayPoints - repmat(newWayPoint,size(wayPoints,1),1)),2));
    [~,newWayPointClosestVertexIndex] = min(sum(abs(wayPoints - repmat(newWayPointClosestVertex,size(wayPoints,1),1)),2));
    edgeWeights(projectionLastWayPointIndex,projectionLastWayPointAIndex) = distanceProjectionLastWayPointToA;
    edgeWeights(projectionLastWayPointAIndex,projectionLastWayPointIndex) = distanceProjectionLastWayPointToA;
    edgeWeights(projectionLastWayPointIndex,projectionLastWayPointBIndex) = distanceProjectionLastWayPointToB;
    edgeWeights(projectionLastWayPointBIndex,projectionLastWayPointIndex) = distanceProjectionLastWayPointToB;
    edgeWeights(newWayPointClosestVertexIndex,newWayPointIndex) = 100;
    edgeWeights(newWayPointIndex,newWayPointClosestVertexIndex) = 100;
    edgeWeights(lastWayPointIndex,projectionLastWayPointIndex) = 1;
    edgeWeights(projectionLastWayPointIndex,lastWayPointIndex) = 1;
    edgeWeights(lastWayPointIndex,projectionLastWayPointAIndex) = 1;
    edgeWeights(projectionLastWayPointAIndex,lastWayPointIndex) = 1;
    edgeWeights(lastWayPointIndex,projectionLastWayPointBIndex) = 1;
    edgeWeights(projectionLastWayPointBIndex,lastWayPointIndex) = 1;
    edgeWeights = sparse(edgeWeights);
    
    %% determine the path to traverse
    wayPointsToTravelIdx = shortestpath(graph(edgeWeights),lastWayPointIndex,newWayPointIndex);
    wayPointsToTravel = wayPoints(wayPointsToTravelIdx,:);
    wayPointsToTravel = prioritySampling(samples,wayPointsToTravel,stepSize,distanceFunction);
end

